/*######     Copyright (c) 1997-2014 Ufasoft  http://ufasoft.com  mailto:support@ufasoft.com,  Sergey Pavlov  mailto:dev@ufasoft.com #########################################################################################################
#                                                                                                                                                                                                                                            #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation;  either version 3, or (at your option) any later version.          #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.   #
# You should have received a copy of the GNU General Public License along with this program; If not, see <http://www.gnu.org/licenses/>                                                                                                      #
############################################################################################################################################################################################################################################*/

#include <el/ext.h>

#include <zlib.h>


namespace Ext {

const size_t ZBUF_SIZE = 2048;

int ZlibCheck(int r) {
	if (r < 0)
		Throw(E_EXT_Zlib + -r);
	return r;
}

#define ZSTM ((z_stream*)m_pimpl)

CompressStream::CompressStream(Stream& stm, CompressionMode mode)
	:	m_stmBase(stm)
	,	Mode(mode)
	,	m_bufpos(0)
	,	m_pimpl(new z_stream)
	,	m_sbuf(ZBUF_SIZE)
	,	m_dbuf(ZBUF_SIZE)
{
}

CompressStream::~CompressStream() {
	if (Mode == CompressionMode::Decompress)
		ZlibCheck(inflateEnd(ZSTM));
	else {
		WritePortion(0, 0, Z_FINISH);
		ZlibCheck(deflateEnd(ZSTM));
	}
	delete ZSTM;
}

void CompressStream::InitImp() const {
	if (Mode == CompressionMode::Decompress) {
		ZeroStruct(*ZSTM);
		ZSTM->next_out = &m_dbuf[0];
		ZSTM->avail_out = m_dbuf.size();
		ZSTM->next_in = &m_sbuf[0];
		ZSTM->avail_in = 0;
		ZlibCheck(inflateInit(ZSTM));
	} else
		Throw(E_NOTIMPL);
}

void CompressStream::ReadPortion() const {
	z_stream& z = *ZSTM;
	int bFlush = 0;
	if (!z.avail_in) {
		z.next_in = &m_sbuf[0];
LAB_MORE_IN:
		for (int b; z.avail_in<m_sbuf.size();) {
			if (bFlush = ((b = m_stmBase.ReadByte()) == -1))
				break;
			z.next_in[z.avail_in++] = (byte)b;
		}
		if (!z.avail_in) {
			m_bufpos = -1;
			return;
		}
	}
	m_bufpos = 0;
	z.next_out = &m_dbuf[0];
	z.avail_out = m_dbuf.size();
	switch (ZlibCheck(inflate(&z, bFlush))) {
	case Z_STREAM_END:
		if (z.next_out == &m_dbuf[0])
			m_bufpos = -1;
		break;
	case Z_OK:
		if (m_bByByteMode) {
			if (z.next_in) {
				m_sbuf.erase(m_sbuf.begin(), m_sbuf.begin()+(z.next_in-&m_sbuf[0]));
				z.next_in = &m_sbuf[0];
			} else
				m_sbuf.resize(1);
		}		
		break;
	case Z_BUF_ERROR:
		if (!z.next_in && m_bByByteMode) {
			m_sbuf.resize(m_sbuf.size()+1);
			goto LAB_MORE_IN;
		}
	default:
		Throw(E_FAIL);
	}
}

bool CompressStream::Eof() const {
	if (!m_bInited) {
		InitImp();
		m_bInited = true;
	}
	if (m_bufpos == -1)
		return true;
	if (&m_dbuf[0]+m_bufpos < ZSTM->next_out)
		return false;
	ReadPortion();
	return m_bufpos == -1;
}

void CompressStream::SetByByteMode(bool v) {
	m_bByByteMode = v;
	m_sbuf.resize(m_bByByteMode ? 1 : ZBUF_SIZE);
}

size_t CompressStream::Read(void *buf, size_t count) const {
	if (Eof())
		Throw(E_EXT_EndOfStream);
	size_t n = std::min(count, size_t(ZSTM->next_out-&m_dbuf[0]-m_bufpos));
	memcpy(buf, &m_dbuf[0]+m_bufpos, n);
	m_bufpos += n;
	return n;
}

void CompressStream::WritePortion(const void *ibuf, size_t count, int flush) {
	if (!m_bInited) {
		InitImp();
		m_bInited = true;
	}
	ZSTM->next_in = (byte*)ibuf;
	ZSTM->avail_in = count;
	while (true) {
		byte buf[4096];
		ZSTM->next_out = buf;
		ZSTM->avail_out = sizeof(buf);
		int r = ZlibCheck(deflate(ZSTM, flush));
		if (int size = sizeof(buf)-ZSTM->avail_out)
			m_stmBase.WriteBuffer(buf, size);
		if ((flush != Z_FINISH || r == Z_STREAM_END) && !ZSTM->avail_in)
			break;
	};
}

void CompressStream::WriteBuffer(const void *buf, size_t count) {
	WritePortion(buf, count, Z_NO_FLUSH);
}

void GZipStream::InitImp() const {
	z_stream& z = *ZSTM;
	ZeroStruct(z);
	z.next_out = &m_dbuf[0];
	z.avail_out = m_dbuf.size();
	z.next_in = &m_sbuf[0];
	z.avail_in = 0;
	if (Mode == CompressionMode::Decompress) {
		ZlibCheck(inflateInit2(&z, 16+MAX_WBITS));
	} else {
		ZlibCheck(deflateInit2(ZSTM, Z_DEFAULT_COMPRESSION, Z_DEFLATED, 31, 8, Z_DEFAULT_STRATEGY));
	}

}

} // Ext::


